#include "testutil.hh"
#include <cxxtest/TestSuite.h>

#include "calibrationrequest.hh"
#include "realposition.hh"

#include <fstream>
#include <sstream>
#include <sys/stat.h>

using namespace std ;

vector<AccessPoint> TestUtil::aps ;
vector<Mobile> TestUtil::mobiles ;
vector<Request*> TestUtil::requests ;
vector<Result> TestUtil::results ;
vector<ReferencePoint> TestUtil::reference_points ;


void TestUtil::tear_down()
{
  aps.clear() ;
  mobiles.clear() ;

  for (vector<Request*>::const_iterator i = requests.begin() ;
       i != requests.end() ; ++i)
    delete *i ;
  requests.clear() ;

  results.clear() ;
  reference_points.clear() ;
}


void TestUtil::set_up()
{
  tear_down() ;
  create_mobile_list() ;
  create_ap_list() ;
  create_request_list() ;
  create_result_list() ;
}

void TestUtil::create_mobile_list()
{
  Mobile mobile1 ;
  mobile1.set_mac_addr("aa:bb:cc:dd:ee:ff") ;
  mobiles.push_back(mobile1) ;
  mobile1.set_mac_addr("aa:bb:cc:dd:ee:77") ;
  mobiles.push_back(mobile1) ;
}

void TestUtil::create_ap_list()
{
  AccessPoint ap1 ;
  ap1.set_mac_addr("11:22:33:44:55:01") ;
  aps.push_back(ap1) ;
  ap1.set_mac_addr("11:22:33:44:55:02") ;
  aps.push_back(ap1) ;
  ap1.set_mac_addr("11:22:33:44:55:03") ;
  aps.push_back(ap1) ;
}

void TestUtil::create_request_list()
{
  // Create measurement list
  vector < unordered_map<string,Measurement> > measurements(3) ;
  Measurement measurement1 ;

  measurement1.add_ss(-58) ;
  measurement1.add_ss(-55) ;
  measurement1.set_ap(&aps[0]) ;
  measurements[0][aps[0].get_mac_addr()] = measurement1 ;
  measurement1.clear() ;
  measurement1.add_ss(-50) ;
  measurement1.set_ap(&aps[2]) ;
  measurements[0][aps[2].get_mac_addr()] = measurement1 ;
  measurement1.clear() ;
  measurement1.add_ss(-42) ;
  measurement1.add_ss(-37) ;
  measurement1.set_ap(&aps[1]) ;
  measurements[0][aps[1].get_mac_addr()] = measurement1 ;

  measurement1.clear() ;
  measurement1.add_ss(-51) ;
  measurement1.add_ss(-19) ;
  measurement1.set_ap(&aps[0]) ;
  measurements[1][aps[0].get_mac_addr()] = measurement1 ;
  measurement1.clear() ;
  measurement1.add_ss(-70) ;
  measurement1.add_ss(-21) ;
  measurement1.set_ap(&aps[1]) ;
  measurements[1][aps[1].get_mac_addr()] = measurement1 ;
  measurement1.clear() ;
  measurement1.add_ss(-59) ;
  measurement1.set_ap(&aps[2]) ;
  measurements[1][aps[2].get_mac_addr()] = measurement1 ;

  measurement1.clear() ;
  measurement1.add_ss(-54) ;
  measurement1.add_ss(-32) ;
  measurement1.set_ap(&aps[0]) ;
  measurements[2][aps[0].get_mac_addr()] = measurement1 ;
  measurement1.clear() ;
  measurement1.add_ss(-45) ;
  measurement1.add_ss(-23) ;
  measurement1.set_ap(&aps[1]) ;
  measurements[2][aps[1].get_mac_addr()] = measurement1 ;
  measurement1.clear() ;
  measurement1.add_ss(-56) ;
  measurement1.set_ap(&aps[2]) ;
  measurements[2][aps[2].get_mac_addr()] = measurement1 ;

  // Create timestamp list
  vector<Timestamp> timestamps ;
  timestamps.push_back(Timestamp(1265120910725)) ;
  timestamps.push_back(Timestamp(1265120911234)) ;
  timestamps.push_back(Timestamp(1265120912345)) ;

  // Create request list
  requests.push_back(new Request(&mobiles[0], timestamps[0],
                                 measurements[0])) ;

  Request request2(&mobiles[1], timestamps[1], measurements[1]) ;
  reference_points.push_back(ReferencePoint(3.2, 4.1, 3.4)) ;
  requests.push_back(new CalibrationRequest(
                       request2, &reference_points.at(0),
                       Direction(2))) ;

  CalibrationRequest *calibration_request =
    dynamic_cast<CalibrationRequest*>(requests.at(1)) ;
  if (calibration_request != NULL)
    calibration_request->reference_point_backward_link() ;

  requests.push_back(new Request(&mobiles[0], timestamps[2],
                                 measurements[2])) ;
}

void TestUtil::create_result_list()
{
  RealPosition algo ;

  for (vector<Request*>::const_iterator i = requests.begin() ;
       i != requests.end() ; ++i)
    {
      CalibrationRequest *calibration_request =
        dynamic_cast<CalibrationRequest*>(*i) ;
      Result result ;
      if (calibration_request != NULL)
        result = algo.compute(*calibration_request) ;
      else
        result = algo.compute(**i) ;
      results.push_back(result) ;
    }
}


// Create the file output_file_name and fill it with the contents of
// output_lines
void TestUtil::fill_file(const string &output_file_name,
                         const vector<string> output_lines)
{
  // Open the file
  std::ofstream output_file ;
  output_file.open(output_file_name.c_str()) ;
  if (! output_file)
    TS_FAIL("Cannot open test file `"+ output_file_name
            +"` for creation!") ;

  // Write contents to the file
  for (std::vector<std::string>::const_iterator i = output_lines.begin() ;
       i != output_lines.end() ; ++i)
    output_file << *i ;

  output_file.close() ;
}


void TestUtil::remove_file(const string &file_name)
{
  if (remove(file_name.c_str()) == -1)
    TS_WARN("Cannot remove test file `"+ file_name +"`!") ;
}


void TestUtil::
create_test_csv_file(const string &file_name, bool with_spaces)
{
  // Fill name and contents of the test CSV file
  vector<string> csv_lines ;
  ostringstream line ;

  if (with_spaces)
    line << "\n   \n     " ;
  line << mobiles[0].get_mac_addr() ;
  line << ';' << requests.at(0)->get_time_sent() ;
  line << ";0;0;0;0;" ;
  line << aps[0].get_mac_addr() ;
  line << ';' << requests.at(0)->get_measurements()
       .at(aps[0].get_mac_addr()).get_ss_list()[0] ;
  line << ';' << aps[2].get_mac_addr() ;
  line << ';' << requests.at(0)->get_measurements()
       .at(aps[2].get_mac_addr()).get_ss_list()[0] ;
  line << ';' << aps[1].get_mac_addr() ;
  line << ';' << requests.at(0)->get_measurements()
       .at(aps[1].get_mac_addr()).get_ss_list()[0] ;
  line << ';' << aps[0].get_mac_addr() ;
  line << ';' << requests.at(0)->get_measurements()
       .at(aps[0].get_mac_addr()).get_ss_list()[1] ;
  line << ';' << aps[1].get_mac_addr() ;
  line << ';' << requests.at(0)->get_measurements()
       .at(aps[1].get_mac_addr()).get_ss_list()[1] ;
  line << '\n' ;
  csv_lines.push_back(line.str()) ;

  line.str("") ;
  line << mobiles[1].get_mac_addr() ;
  CalibrationRequest *calibration_request =
    dynamic_cast<CalibrationRequest*>(requests.at(1)) ;
  assert(calibration_request) ;
  line << ';' << calibration_request->get_time_sent() ;
  line << ';' << reference_points.at(0).get_x() ;
  line << ';' << reference_points.at(0).get_y() ;
  line << ';' << reference_points.at(0).get_z() ;
  line << ';' << static_cast<int>(calibration_request->get_direction()) ;
  line << ';' << aps[2].get_mac_addr() ;
  line << ';' << calibration_request->get_measurements()
       .at(aps[2].get_mac_addr()).get_ss_list()[0] ;
  line << ';' << aps[0].get_mac_addr() ;
  line << ';' << calibration_request->get_measurements()
       .at(aps[0].get_mac_addr()).get_ss_list()[0] ;
  line << ';' << aps[1].get_mac_addr() ;
  line << ';' << calibration_request->get_measurements()
       .at(aps[1].get_mac_addr()).get_ss_list()[0] ;
  line << ';' << aps[1].get_mac_addr() ;
  line << ';' << calibration_request->get_measurements()
       .at(aps[1].get_mac_addr()).get_ss_list()[1] ;
  line << ';' << aps[0].get_mac_addr() ;
  line << ';' << calibration_request->get_measurements()
       .at(aps[0].get_mac_addr()).get_ss_list()[1] ;
  line << '\n' ;
  if (with_spaces)
    line << "     \n\
\n\t\n\
\n   \t   \n\
\n" ;
  csv_lines.push_back(line.str()) ;

  line.str("") ;
  if (with_spaces)
    line << '\t' ;
  line << mobiles[0].get_mac_addr() ;
  line << ';' << requests.at(2)->get_time_sent() ;
  line << ";0;0;0;0;" ;
  line << aps[2].get_mac_addr() ;
  line << ';' << requests.at(2)->get_measurements()
       .at(aps[2].get_mac_addr()).get_ss_list()[0] ;
  line << ';' << aps[1].get_mac_addr() ;
  line << ';' << requests.at(2)->get_measurements()
       .at(aps[1].get_mac_addr()).get_ss_list()[0] ;
  line << ';' << aps[0].get_mac_addr() ;
  line << ';' << requests.at(2)->get_measurements()
       .at(aps[0].get_mac_addr()).get_ss_list()[0] ;
  line << ';' << aps[1].get_mac_addr() ;
  line << ';' << requests.at(2)->get_measurements()
       .at(aps[1].get_mac_addr()).get_ss_list()[1] ;
  line << ';' << aps[0].get_mac_addr() ;
  line << ';' << requests.at(2)->get_measurements()
       .at(aps[0].get_mac_addr()).get_ss_list()[1] ;
  line << '\n' ;
  if (with_spaces)
    line << "\n\t\n" ;
  csv_lines.push_back(line.str()) ;

  // Create and fill the test CSV file
  fill_file(file_name, csv_lines) ;
}


// Test equality of two Request comparing pointed values instead of
// pointers and taking care of the unordered_map
bool TestUtil::request_equals(const Request &first,
                              const Request &second)
{
  // Try a classical comparison
  if (first == second)
    return true ;

  // Compare timestamp
  if (first.get_time_sent() != second.get_time_sent())
    return false ;

  // Compare mobile values
  if (first.get_mobile() == NULL || second.get_mobile() == NULL)
    {
      if (first.get_mobile() != NULL || second.get_mobile() != NULL)
        return false ;
    }
  else if (*first.get_mobile() != *second.get_mobile())
    return false ;

  // Compare measurements
  if (first.get_measurements() != second.get_measurements())
    if (! measurements_unordered_map_equals(first.get_measurements(),
                                            second.get_measurements()))
      return false ;

  return true ;
}


// Test equality of two unordered_map<string,Measurement>
bool TestUtil::
measurements_unordered_map_equals(
  unordered_map<string,Measurement> first,
  unordered_map<string,Measurement> second)
{
  if (first.size() != second.size())
    return false ;

  // For each element in 'first', we look for the same element
  // in 'second'
  for (unordered_map<string,Measurement>::const_iterator i =
         first.begin() ; i != first.end() ; ++i)
    {
      unordered_map<string,Measurement>::const_iterator
      second_measurement = second.find(i->first) ;
      if (second_measurement == second.end())
        return false ;
      if (! measurement_equals(i->second, second_measurement->second))
        return false ;
    }

  return true ;
}


// Test equality of two Measurement comparing pointed values instead of
// pointers
bool TestUtil::measurement_equals(const Measurement &first,
                                  const Measurement &second)
{
  // Try a classical comparison
  if (first == second)
    return true ;

  // Compare average_dbm
  if (first.get_average_dbm() != second.get_average_dbm())
    return false ;

  // Compare ap values
  if (first.get_ap() == NULL || second.get_ap() == NULL)
    {
      if (first.get_ap() != NULL || second.get_ap() != NULL)
        return false ;
    }
  else if (*first.get_ap() != *second.get_ap())
    return false ;

  // Compare ss_list
  if (first.get_ss_list() != second.get_ss_list())
    return false ;

  return true ;
}


bool TestUtil::file_size_equals(const string &file1_name,
                                const string &file2_name)
{
  try
    {
      return (file_size(file1_name) == file_size(file2_name)) ;
    }
  catch (...)
    {
      return false ;
    }
}


off_t TestUtil::file_size(const string &file_name)
{
  struct stat file_info ;
  if (stat(file_name.c_str(), &file_info) != 0)
    throw "Cannot get information for file `" + file_name + "`!" ;
  return file_info.st_size ;
}
